% LoMeshAveraging (subclass of MeshProlongation) Prolongate lowest order
%   nonconforming CR finite element function to refined mesh by averaging 
%   in the vertices.
%
%   P = LoMeshAveraging(fes) returns a handle to the prolongation object
%       associated to the nonconforming finite element space fes. The prolongation 
%       matrix P.matrix is set automatically at mesh refinement.
%
% See also Prolongation, MeshProlongation

classdef LoMeshQuasiInterpolation < MeshProlongation
    %% properties
    properties (Access=protected)
        feType
    end

    %% methods
    methods (Access=public)
        function obj = LoMeshQuasiInterpolation(fes)
            obj = obj@MeshProlongation(fes);
            switch class(fes.finiteElement)
                case 'LowestOrderCRFe'
                    obj.feType = 'CR';
                otherwise
                    eid = 'LoMeshQuasiInterpolation:wrongFeType';
                    msg = 'LoMeshQuasiInterpolation needs a lowest order CR finite element space.';
                    throwAsCaller(MException(eid, msg));
            end
        end
    end
    
    methods (Access=protected)
        function setupMatrix(obj, mesh, data)
            if isequal(obj.feType, 'CR')
                setupMatrixCR(obj, mesh, data);
            end
        end

        % override only in case of CR elements
        function connectDofs(obj, mesh, data)
            if isequal(obj.feType, 'CR')
                connectDofsCR(obj, mesh, data);
            end
        end

        function setupMatrixCR(obj, ~, data)
            % use that for lowest order CR elements the dofs correspond to
            % coordinates and new coordinates reside on edges or on inner nodes
            %
            % fixed boundary dofs remain zero 

            % determine dofs of CR function space
            dofs = getDofs(obj.fes);

            % numbers of bisected edges
            nUnbisectedEdges = nnz(~data.bisectedEdges);
            nBisectedEdges = nnz(data.bisectedEdges);

            % determine number of dofs in fine CR function space:
            %   nEdges + 1 for each bisected edge + number of new edges
            nNewDofs = dofs.nDofs + nBisectedEdges ...
                       + sum(data.nInnerEdges .* data.nRefinedElements);

            % old-new relation of edge indices
            %   The following vector contains the indices of all old edges
            %   in the new edges list. For every refined edge, the given
            %   index contains the index of the first child edge. It
            %   follows the index of the second child edge directly after
            %   it.
            %   
            %   See oldEdge2parent in updateData function
            oldEdge2children = cumsum([1; 1+data.bisectedEdges]);

            % initialize components of sparse matrix generation
            [I, J, V] = deal(zeros(nUnbisectedEdges + 2*nBisectedEdges, 1));

            % Copy dofs for unbisected edges
            I(1:nUnbisectedEdges) = find(~data.bisectedEdges);
            J(1:nUnbisectedEdges) = oldEdge2children(~data.bisectedEdges);
            V(1:nUnbisectedEdges) = 1;
            idxNr = nUnbisectedEdges;  % number of unbisected edges

            % Average dofs of children edges to obtain parent value
            idx = idxNr + (1:2*nBisectedEdges);
            I(idx) = repelem(find(data.bisectedEdges), 2, 1);
            J(idx) = repelem(oldEdge2children(data.bisectedEdges), 2, 1) ...
                     + repmat([0;1], nBisectedEdges, 1);
            V(idx) = 0.5;

            % assemble quasi-interpolation matrix and store its transpose
            % as prolongation matrix
            QuasiInterpolation = sparse(I, J, V, dofs.nDofs, nNewDofs);
            obj.matrix = transpose(QuasiInterpolation);
        end

        function connectDofsCR(~, ~, ~)
            % relax
        end
    end
end
